'''Contains classes and functions to handle spatial queries'''
from io import BytesIO
import pickle
from typing import Collection
import shapely
import shapefile
import os
from shapely.geometry import shape
from shapely.strtree import STRtree
from shapely.geometry import Polygon
from shapely import wkt
import pygeoif
import pickle
from pkg_resources import resource_filename

from creaap.spatial.utils import convert_goem
from collections.abc import Iterable

class SpatialMatcher:
    '''Allows you to perform spatial queries over a set of geometries
    This objects "reasons" with WKT geometries, if you don't know what
    WKT is, check this out:
    https://en.wikipedia.org/wiki/Well-known_text_representation_of_geometry 
    
    Attributes
    ----------
    index: shapely.geometry.STRtree
        the internal spatial index. Don't fuck with it.

    g: list of shapely.geometry.base.BaseGeometry
        the plottable representation of stored geometries

    labels: list of string
        the labels associated to each geometry object

    '''
    
    def __init__(self):
        '''creates a new, empty, object. No spatial index is built, build it with
        the build_spatial_index method or load it from file/stram with load/loads method'''
        self.index = None
        self.g = []
        self.labels =[]
          
    def build_spatial_index(self, shapefile = None, geometries = None, labels=None):
        '''Builds the spatial index. Takes as input either a shapefile or a list of geometries
        and indexes them allowing to perform queries on such elements.
        Each geometry object should have an associated label that identifies it that can be
        provided either embedded in the shapefile or in a separated list.

        Parameters
        ----------
        shapefile: shapefile.Reader, default None
            a shepefile object containiang the geometries you'd like to index

        geometries: list of shapely.geometry.base.BaseGeometry or WKT strings, defualt None
            a list of Shapely geometries or WKT strings in the likes of 'POINT(11.01 42.67)'

        labels: string or list of string
            either the shapefile attribute you want to use as label, or an actual list of string labels;
            if you provide labels as a list, its lenght must match the number of indexed geometry objects

        Returns
        -------
        None
        
        '''
        self.g=[]
        if shapefile:
            for s in shapefile.shapes():
                self.g.append(shape(s))
        else:
            if geometries is not None:
                for gm in geometries:
                    if isinstance(gm, str):
                        try: 
                            self.g.append(wkt.loads(gm))
                        except:
                            raise ValueError("geometries must be valid WKT strings, found " + str(gm))
                    elif isinstance(gm, shapely.geometry.base.BaseGeometry):
                        # just shove the bastard in
                        self.g.append(gm)
                    else:
                        raise TypeError("geometries must be either WKT strings or Shapely geometries, found " + str(type(gm)))
        # build the index
        self.index = STRtree(self.g)
        if isinstance(labels, str):
            for record in shapefile.records():
                self.labels.append(record.as_dict()[labels])
        elif isinstance (labels, list):
            self.labels = labels
        elif isinstance(labels, Iterable) or isinstance(labels, Collection):
            self.labels = list(labels)
        else:
            raise TypeError("labels can be either a field in the Shapefile or a list of values")
              

    def get_intersection(self, geometry_wkt):
        '''
        Returns the labels of the indexed geometries which intesrect the query geometry

        Parameters
        ----------
        geometry_wkt: string
            a WKT representation of the query geometry

        Returns
        -------
        list
            labels of intersecting geometries
        '''
        query_geom =wkt.loads(geometry_wkt)
        result = self.index.query(query_geom)
        out_idxs=[]
        for r in result:
            res_geom = self.g[r]
            if res_geom.intersects(query_geom):
                out_idxs.append(r)
        return [self.labels[i] for i in out_idxs]
    
    def get_intersection_percentages(self, geometry_wkt, areas=None):
        '''
        Returns the labels of the indexed geometries which intesrect the query geometry
        and their intersection percentages.

        Parameters
        ----------
        geometry_wkt: string
            a WKT representation of the query geometry.

        areas: list of strings, default None
            a filtering paramter, allows to restrict the query to a subset of the index 
            by specifiying the set of labels of interest; if left to None, the query 
            will run on all indexed geometries.
       
        Returns
        -------
        dict
            a label -> percentage dictionary; the percentage value represents
            how much of the geometry identified by the label intersercts with 
            the query geometry.
        '''

        if areas is None:
            areas = self.labels
        target_shape = wkt.loads(geometry_wkt)
        target_area = target_shape.area
        out = {}
        for l in areas:
            intersection=target_shape.intersection(self.g[self.labels.index(l)])
            out[l] = intersection.area / target_shape.area 
        return out
    
    def get_nearest_geom(self, geometry_wkt):
        '''
        Returns the label of the nearest indexed geometry

        Parameters
        ----------
        geometry_wkt: string
            a WKT representation of the query geometry

        Returns
        -------
        string
            label of the nearest geometry
        '''
        query_geom =wkt.loads(geometry_wkt)
        result = self.index.nearest(query_geom)
        return self.labels[result]
    
    def save(self, path):
        '''serializes the SpatialMatcher instance to a file
        
        Parameters
        ----------
        path: string
            the fielsystem location of the destination file

        Returns
        -------
        None
        '''
        file = open(path, 'wb')
        pickle.dump((self.index, self.g, self.labels), file)
        file.close()
    
    def load(self, path):
        '''loads the SpatialMatcher's index and labels from a file
        
        Parameters
        ----------
        path: string
            the fielsystem location of the target file

        Returns
        -------
        None
        '''
        file = open(path, 'rb')
        self.index, self.g, self.labels = pickle.load(file)
        file.close()
    
    def dumps(self):
        '''
        serializes the SpatialMatcher instance to a bytes object
        
        Parameters
        ----------
        None

        Returns
        -------
        bytes
            all the relevant data in the current SpatialMatcher instance
            pacakged in a nice byte string
        '''
        return pickle.dumps((self.index, self.g, self.labels))
    
    def loads(self, stream):
        '''loads the SpatialMatcher's index and labels from a bytes object
        
        Parameters
        ----------
        stream: bytes
            the byte object that contains the SpatialMatcher's index and labels

        Returns
        -------
        None
        '''
        self.index, self.g, self.labels = pickle.loads(stream)


class NoNutsError(Exception):
    """Exception raised when you are totally out of NUTs boundaries.

    Attributes
    ----------
    message: string
        Explanation of the error
    """

    def __init__(self, message="I came to get NUTs and raise exceptions, and I am totally out of NUTs"):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.salary} -> {self.message}'

def get_nuts_matcher(level = 0):
    '''Loads a pre-computed NUTS matcher. Quick and dirty, expect little detail.

    Parameters
    ----------
    level: int in range 0-3
        The NUTS level you want to load, where 0 is nations,
        1 is macro-regions (e.g. Central Italy), 2 is 
        administrative regions (e.g. Lazio), and 3 is districts
        (e.g. Rome) 

    '''
    try:
        matcher = SpatialMatcher()
        matcher.load(resource_filename(__name__, 'data' + os.sep + 'nuts' + str(level)))
        return matcher
    except:
        raise NoNutsError('No level '+ str(level) + ' NUTS boundaries. Please use 0, 1, 2, or 3.')


def _get_nuts_level(nuts_shapefile, nuts_level = 0, attribute = 'NAME_LATN'):
    '''Helper function that extracts a NUTs level geography from a shapefile'''
    shapes =[]
    names = []
    all_records = nuts_shapefile.records() # this rat bastard is a labelled list of some sort that looks like a dictionary but ain't a dictionary
    all_shapes = nuts_shapefile.shapes() # and this is a Shapes object which is really a list
    for idx in range(len(all_shapes)):
        record = all_records[idx]
        if record.as_dict()['LEVL_CODE'] == nuts_level:
            names.append(record.as_dict()[attribute].rstrip('\x00'))
            shapes.append(all_shapes[idx])
    return shapefile.Shapes(shapes), names


def build_nuts_matcher(nuts_shapefile, nuts_level = 0, input_projection = "epsg:4326",output_projection = "epsg:4326", attribute = 'NAME_LATN'):
    '''
    Converts the official EU NUTs shapefie into a practical matcher.

    Get you NUTS shapefile here
    https://ec.europa.eu/eurostat/web/gisco/geodata/reference-data/administrative-units-statistical-units/nuts
    ad find a way to unzip it and load it (spoiler: look for read_shape_from_zip in creaap.spatial.utils)

    Parameters
    ----------
    nuts_shapefile: a shapefile object
    '''
    shapes, names = _get_nuts_level(nuts_shapefile, nuts_level, attribute) # get correct level shapefile
    shp , _ , dbf = convert_goem(shapes,input_projection, output_projection, names, flip_coords=False)
    wgs84_shapes = shapefile.Reader(shp=shp, dbf=dbf)
    matcher = SpatialMatcher()
    matcher.build_spatial_index(wgs84_shapes, labels = names)
    return matcher



'''Classes and functions to upload and download stuff from Azure Blob containers'''
from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient
from azure.core.exceptions import ResourceExistsError
import sys, inspect
import os
import logging


class AzureBlobClient:
    '''A simple client to perform upload and download operations from blob storage'''
    def __init__(self, connection_string, container_name):
        '''Creates a new connection client for a Blob Container.

        Parameters
        ----------
        conenction_string: string
            the Azure Storage connection string. Grab it in the Azure Portal
        
        container_name: string
            the name of the container you want to connect to
        '''
        service_client = BlobServiceClient.from_connection_string(connection_string)
        self.client = service_client.get_container_client(container_name)

    def upload(self, source, dest):
        '''Upload a file or directory to a path inside the container
        Use this method if you don't know what the source path points at.

        Parameters
        ----------
        source: string
            file system path of the target resource, can be a file or a directory
        
        dest: string
            blob path for the uploaded resouces

        Returns
        -------
        None
        '''
        if (os.path.isdir(source)):
            self.upload_dir(source, dest)
        else:
            self.upload_file(source, dest)

    def upload_file(self, source, dest):
        '''
        Upload a single file to a path inside the container

        Parameters
        ----------
        source: string
            file system path of the target resource
        
        dest: string
            blob path for the uploaded file

        Returns
        -------
        None
        '''
        logging.info(f'Uploading {source} to {dest}')
        with open(source, 'rb') as data:
            self.upload_stream(data=data, dest=dest)

    def upload_stream(self, data, dest):
        '''
        Upload a Bytes stream to a path inside the container

        Parameters
        ----------
        data: bytes
            the resource you want to upload, represented as a bytes object
        
        dest: string
            blob path for the uploaded file

        Returns
        -------
        None
        '''
        self.client.upload_blob(name=dest, data=data, overwrite=True)

    def upload_dir(self, source, dest):
        '''
        Upload a directory to a path inside the container

        Parameters
        ----------
        source: string
            file system path of the target resource
        
        dest: string
            blob path for the uploaded resouces

        Returns
        -------
        None
        '''
        prefix = '' if dest == '' else dest + '/'
        prefix += os.path.basename(source) + '/'
        for root, _, files in os.walk(source):
            for name in files:
                dir_part = os.path.relpath(root, source)
                dir_part = '' if dir_part == '.' else dir_part + '/'
                file_path = os.path.join(root, name)
                blob_path = prefix + dir_part + name
                self.upload_file(file_path, blob_path)

    def download(self, source, dest):
        '''
        Download a file or directory to a path on the local filesystem.
        Use this method if you are not sure if the blob path points at
        a file or a directory.

        Parameters
        ----------
        source: string
            blob path of the target resource
        
        dest: string
            file system path of the output directory/file

        Returns
        -------
        None
        '''
        if not dest:
            raise Exception('A destination must be provided')

        blobs = self.ls_files(source, recursive=True)
        if blobs:
        # if source is a directory, dest must also be a directory
            if not source == '' and not source.endswith('/'):
                source += '/'
            if not dest.endswith(os.sep):
                dest += os.sep
        # append the directory name from source to the destination
        dest += os.path.basename(os.path.normpath(source)) + os.sep

        blobs = [source + blob for blob in blobs]
        for blob in blobs:
            blob_dest = dest + os.path.relpath(blob, source)
            self.download_file(blob, blob_dest)
        else:
            # in this branch we are delaing witha  single file
            self.download_file(source, dest)
        
    def download_file(self, source, dest):
        '''
        Download a single file to a path on the local filesystem.

        Parameters
        ----------
        source: string
            blob path of the target resource
        
        dest: string
            file system path of the output file

        Returns
        -------
        None
        '''
        # dest is a directory if ending with '/' or '.', otherwise it's a file
        if dest.endswith('.'):
            dest += '/'
            blob_dest = dest + os.path.basename(source) if dest.endswith('/') else dest
        logging.info(f'Downloading {source} to {blob_dest}')
        os.makedirs(os.path.dirname(blob_dest), exist_ok=True)
        bc = self.client.get_blob_client(blob=source)
        with open(blob_dest, 'wb') as file:
            data = bc.download_blob()
            file.write(data.readall())
    
    def download_to_bytes(self, source):
        '''download a single blob to a byte array.
        
        Parameters
        ----------
        source: string
            blob path of the target resource

        Returns
        -------
        bytes
            byte-object representation of the blob resource
        '''
        logging.info(f'Downloading {source} to byte array')
        bc = self.client.get_blob_client(blob=source)
        data = bc.download_blob()
        return data.readall()

    def ls_files(self, path, recursive=False):
        '''
        List files under a path, optionally recursively

        Parameters
        ----------
        path: string
            the base path in the container to ls

        recursive: bool, default False
            whether or not to go recursive on "directories" and get
            the whole file tree

        Returns
        generator
            strings that represent items in the container
        '''
        if not path == '' and not path.endswith('/'):
            path += '/'

        blob_iter = self.client.list_blobs(name_starts_with=path)
        for blob in blob_iter:
            relative_path = blob.name.lstrip(path)
            if recursive or not '/' in relative_path:
                yield blob.name

    def ls_dirs(self, path, recursive=False):
        '''
        List directories under a path, optionally recursively.
        Mind that in Azure Blob Storage directories are not "real" directories
        like in a file system, but rather prefeixes you use to better organize
        your blobs.

        Parameters
        ----------
        path: string
            the base path in the container to ls

        recursive: bool, default False
            whether or not to go recursive on "directories" and fetch
            all sub-subdirectories as well

        Returns
        generator
            strings that represent "directories" in the container
        -------
        '''
        if not path == '' and not path.endswith('/'):
            path += '/'

        blob_iter = self.client.list_blobs(name_starts_with=path)
        dirs = set()
        for blob in blob_iter:
            # here we just have to strip the filename
            relative_dir = '/'.join(blob.name.split('/')[:-1])
            if relative_dir and (recursive or not '/' in relative_dir) and not relative_dir in dirs:
                yield relative_dir
                dirs.add(relative_dir)
    
    def delete_file(self,path):
        '''
        Delete a single file inside the container

        Parameters
        ----------
        path: string
            blob path for the file to delete
        
        Returns
        -------
        None
        '''
        bc = self.client.get_blob_client(blob=path)
        bc.delete_blob()


import re
import json
from sqlite3 import Timestamp
import numpy as np
from datetime import date, timedelta, datetime, timezone
import pandas as pd
import pytz

AZURE_TABLES_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

def read_json_from_bytes(bytes):
	'''converts a bytes array into a Json'''
	my_json = bytes.decode('utf8')
	return json.loads(my_json)

# DATETIME utilities
def to_datetime(date_object):
    '''
    Swiss-army knife function to read datetimes.
    It tries its best to convert anythinkg into a datetime object in the likes of this: 2021-05-19 02:00:00+02:00
    this does not deal with timezones, if your datetime object/string is naive, then the result will be naive too
    
    Parameters
    ----------
    date_object: whatever
        whatever, like, really anbyting goes. If there's a datetime in there, we'll get it

    Returns
    -------
    datetime
        its best effort at parsing the fuckery in date_object
    
    '''
    if isinstance(date_object, pd.Timestamp):
        return timestamp_to_datetime(date_object)
    if isinstance(date_object, datetime):
        return date_object
    elif isinstance(date_object, str):
        try:
            return timestamp_to_datetime(pd.to_datetime(date_object))
        except:
            raise ValueError('Unable to parse ' + str(date_object) + ' into a datetime object.')

def to_tz_aware_datetime(date_object, timezone=pytz.utc, fallback = None):
    '''read dates and fails silently, returns a TZ aware datetime. Good for parsing anything that
    may be a datetime, like an input parameter or whaterver shit the storage throws at you
    
    Parameters
    ----------
    date_object: wahtever
        an object you think it can be interpreted as a datetime
    timezone:pytz timezone OR str, 
        a timezone from the pytz package, defaults to pytz.utc; important: this parameter is relevant
        only if the passed datetime is naive (i.e. no timezone information), otherwise the timezone information
        in the date_string argument will be kept.
    fallback: anything
        what to return if anything fails. If None, exceptions will be raised, otherwise it will fail silently.

    Returns
    -------
    datetime
        its best effort at parsing the fuckery in date_object
  
    '''
    try:
        if isinstance(timezone, pytz.BaseTzInfo):
            return denaive_datetime(to_datetime(date_object), timezone)
        else:
            return denaive_datetime(to_datetime(date_object), pytz.timezone(timezone))  
    except  Exception as ex:
        # time to fail silently
        if fallback is not None:
            return fallback
        else:
            # or to yell at failure
            raise ex

def timestamp_to_datetime(timestamp, fallback=None):
    '''converts Pandas timestamp into a datetime.datetime object'''
    if isinstance(timestamp, Timestamp):
        try:
            out = timestamp.to_pydatetime()
            if pd.isnull(out):
                return fallback
            return out
        except:
            # fail silently
            return fallback
    # if it's aready ok do nothing
    elif isinstance(timestamp, datetime):
        return timestamp
    
   
    
def denaive_datetime(d, timezone = pytz.utc):
    '''turns a naive datime into a timezone-aware one, if the object is already TZ aware, it does nothing'''
    if (d.tzinfo is None) or (d.tzinfo.utcoffset(d) is None):
        return d.replace(tzinfo=timezone)
    else:
        # in this branch the datetime object is already TZ-aware hence do nothing
        return d
    

def datetime_to_tables_string(date_object):
    return date_object.strftime(AZURE_TABLES_DATETIME_FORMAT)


# GEO formats utils

def get_lat_lon_from_wkt(geom_string):
  # we have to parse this POINT (Lon Lat)
  try:
    m = re.search(r'(\d+\.?\d*)\s+(\d+\.?\d*)', geom_string)
    lon = m.group(1)
    lat = m.group(2)
    return lon, lat
  except:
    return None, None

def unfuck_numpy_formats(object):
    '''values that come out of NetCDF files are not in native Python types, but wrapped in numpy weirdness.
    We have to fix it otherwise Azure Tables will throw weird ass errors about malformed tuples'''
    if isinstance(object, np.generic): 
        return object.item()
    else:
        return object



'''A collection of functions to interact with Azure Active Directory'''
import base64
import logging
import requests
import msal
import os
import json
import jwt
import logging


from datetime import datetime


# FIXED Microsoft-dictated endpoints and parameters
graphURI = 'https://graph.microsoft.com'
scope = ['https://graph.microsoft.com/.default']

# AAD Application parameters
# when you add an AAD authentication to your function app, it will create an AAD Application
# go to its page in the Portal, edit it by adding user roles, then add the following permissions:
'''
Application.Read.All
Directory.Read.All
User.Read
User.Read.All
User.ReadBasic.All
'''
# save the whole thing (some app premissions may require approval from your AD Admin).
# Now you can fill properly the following fields
# you can take them from the APPLICATION MANIFEST file
# it looks like '32cc535d-3164-4d2c-99e4-92324ada27b3', it's the "appId" in manifest
# look into the "appRoles" object and pick "id" as key and "value" as value
def _get_client_id_from_env():
	try:
		clientID = os.environ['MICROSOFT_PROVIDER_AUTHENTICATION_CLIENT_ID']
	except KeyError:
		clientID = None
		logging.warning('[WARNING] AAD Client ID not found in environment variables, pass it explicitly with the clientID parameter')
	return clientID

# CLIENT SECRET
# you have to generate this one from the Certificates & Secretes blade. Since secrets last
# 2 years tops, let's slap it into a CONFIGURATION PARAMETER so you won't have to re-deploy
# the function app every fucking time the little motherfucker expires.
def _get_client_secret_from_env():
	try:
		clientSecret = os.environ['MICROSOFT_PROVIDER_AUTHENTICATION_SECRET']
	except KeyError:
		clientSecret = None
		logging.warning('[WARNING] AAD Client secret not found in environment variables, pass it explicitly with the clientSecret parameter')
	return clientSecret

# this is a rat bastard https://stackoverflow.com/questions/37151346/authorization-identitynotfound-error-while-accessing-graph-api
# it looks like this '94dc0eeb-2c10-4dbf-a033-c0683ba0eec9' 
# you find this in the Overvew blade in the AAD application
def _get_tenant_ID_from_env():
	try:
		tenantID = os.environ['MICROSOFT_PROVIDER_AUTHENTICATION_TENANT_ID']
	except KeyError:
		tenantID = None
		logging.warning('[WARNING] AAD Tenant ID not found in environment variables, pass it explicitly with the tenantID parameter')
	return tenantID

# ENTERPRISE APPLICATION parameters
# now move to the Enterpise Application object (is should have the same name as the AAD application and it
# should be accessible from the "overview" blade in the AAD application)
# its ID looks like this '12da6679-c0d1-43fe-9c29-56965dffadf2'
# and its hideous!
def _get_object_id_from_env():
	try:
		object_id = os.environ['MICROSOFT_PROVIDER_AUTHENTICATION_APPLICATION_OBJECT_ID']
	except KeyError:
		object_id = None
		logging.warning('[WARNING] AAD Enterprise Application ID not found in environment variables, pass it explicitly with the object_id parameter')
	return object_id

# HANDLING AUTHENTICATION
def msgraph_auth(
	clientID = None,
	clientSecret = None,
	tenantID = None,
	base_authority = 'https://login.microsoftonline.com/'):
    '''gets an appropriate request header to include in all subsequent AAD requests.
	It includes some caching, so feel free to call as many times as you like'''
    if clientID is None:
        clientID = _get_client_id_from_env()
    if clientSecret is None:
        clientSecret = _get_client_secret_from_env()
    if tenantID is None:
        tenantID = _get_tenant_ID_from_env()
    authority = base_authority + tenantID
    app = msal.ConfidentialClientApplication(clientID, authority=authority, client_credential = clientSecret)
    try:
        accessToken = app.acquire_token_silent(scope, account=None)
        if not accessToken:
            try:
                accessToken = app.acquire_token_for_client(scopes=scope)
                if accessToken['access_token']:
                    print('New access token retreived')
                    requestHeaders = {'Authorization': 'Bearer ' + accessToken['access_token']}
                else:
                    print('Error aquiring authorization token. Check your tenantID, clientID and clientSecret.')
            except:
                pass 
        else:
            print('Token retreived from MSAL Cache')
            requestHeaders = {'Authorization': 'Bearer ' + accessToken['access_token']}
        return requestHeaders
    except Exception as err:
        print(err)

# QUERYING MS GRAPH
def msgraph_request(resource,requestHeaders = None, **kwargs):
	# Request
	if requestHeaders is None:
		requestHeaders = msgraph_auth(**kwargs)
	results = requests.get(resource, headers=requestHeaders).json()
	return results

# QUERY FORMATTING METHODS
# get a user's information
# AAD REST API https://docs.microsoft.com/en-us/graph/api/user-get?view=graph-rest-1.0&tabs=http
def get_user_anagraphics(uid, requestHeaders = None, **kwargs):
	request_url = 'https://graph.microsoft.com/v1.0/users/'+str(uid)
	return msgraph_request(request_url, requestHeaders, **kwargs)

def get_user_application_roles(uid, requestHeaders=None, **kwargs):
	request_url = 'https://graph.microsoft.com/v1.0/users/' + str(uid)+ '/appRoleAssignments'
	return msgraph_request(request_url, requestHeaders, **kwargs)

def get_users(requestHeaders = None, **kwargs):
	request_url = 'https://graph.microsoft.com/v1.0/users/'
	return msgraph_request(request_url, requestHeaders, **kwargs)

def get_application_users(requestHeaders = None, object_id = None , **kwargs):
    if object_id is None:
        object_id = _get_object_id_from_env()
    request_url = 'https://graph.microsoft.com/v1.0/servicePrincipals/'+ object_id +'/appRoleAssignedTo'
    return msgraph_request(request_url, requestHeaders, **kwargs)
# fetches objects like this
#{
# 'id': 'MLDDJlTnckqtOAx9HxAwnEdirTTyNfNKtnlr0vQ4RCw', 
# 'deletedDateTime': None, 
# 'appRoleId': 'd1717ff7-5dcf-4548-b8d8-80873b25c819', 
# 'createdDateTime': '2021-10-25T09:31:32.6833212Z', 
# 'principalDisplayName': 'dario de nart', 
# 'principalId': '26c3b030-e754-4a72-ad38-0c7d1f10309c', 
# 'principalType': 'User', 
# 'resourceDisplayName': 'crea-bee-hive-management-backend', 
# 'resourceId': '12da6679-c0d1-43fe-9c29-56965dffadf2'
# }


### FINALLY: getting the user's data from authentication

def read_user_authorization(request, **kwargs):
	'''Get user data from the http request object'''
	try:
		return read_user_from_jwt(request, **kwargs)
	except AuthorizationError as ae:
		try: 
			return parse_client_principal(request.headers['x-ms-client-principal'])
		except KeyError as ke:
			# no client principal at all, let's show to the client the JWT authentication error
			raise ae
		except  Exception as e:
			# there is a client principal, hence we show the client principal parsing error
			raise AuthorizationError({"code": 'Bad client principal',"description": str(e)}, code = 401)

# https://docs.microsoft.com/en-us/azure/active-directory/develop/scenario-daemon-acquire-token?tabs=python#acquiretokenforclient-api
'''
POST /{tenant}/oauth2/v2.0/token HTTP/1.1           //Line breaks for clarity
Host: login.microsoftonline.com
Content-Type: application/x-www-form-urlencoded

client_id=535fb089-9ff3-47b6-9bfb-4f1264799865
&scope=https%3A%2F%2Fgraph.microsoft.com%2F.default
&client_secret=sampleCredentia1s
&grant_type=client_credentials
'''
# https://blog.darrenjrobinson.com/microsoft-graph-using-msal-with-python/

def parse_client_principal(principal):
	'''given the b64 representation of the client principal, this function
	returns a dictionary with the relevant information'''
	cp_dict =  json.loads(base64.b64decode(principal).decode("utf-8"))
	if cp_dict['auth_typ'] == 'aad':
		claim_dict ={}
		for claim in cp_dict['claims']:
			claim_dict[claim['typ']] = claim['val']
		# now 
		return {
			'uid': claim_dict.get('http://schemas.microsoft.com/identity/claims/objectidentifier'),
			'name' :claim_dict.get('preferred_username'),
			'display_name' : claim_dict.get('name'),
			'email' : claim_dict.get('http://schemas.xmlsoap.org/ws/2005/05/identity/claims/emailaddress'),
			'role' : claim_dict.get('roles')
		}
	return None


def read_user_from_jwt(request, **kwargs):
	'''reads the JWT object from client request'''
	token = get_token_auth_header(request)
	decoded = jwt.decode(token, options={"verify_signature": False})
	return get_user_anagraphics(decoded.get("oid"), **kwargs)



def get_token_auth_header(request):
    """Obtains the Access Token from the Authorization Header
    """
    auth = request.headers.get("Authorization", None)
    if not auth:
        raise AuthorizationError({"code": "authorization_header_missing",
                        "description":
                            "Authorization header is expected"}, 401)

    parts = auth.split()

    if parts[0].lower() != "bearer":
        raise AuthorizationError({"code": "invalid_header",
                        "description":
                            "Authorization header must start with"
                            " Bearer"}, 401)
    elif len(parts) == 1:
        raise AuthorizationError({"code": "invalid_header",
                        "description": "Token not found"}, 401)
    elif len(parts) > 2:
        raise AuthorizationError({"code": "invalid_header",
                        "description":
                            "Authorization header must be"
                            " Bearer token"}, 401)

    token = parts[1]
    #print(token)
    #logging.info(str(token))
    return token

class AuthorizationError(Exception):
	'''autorization error to show in case of user auth clusterfuck'''
	def __init__(self, message, code):
		self.message=message
		self.code = code


import random
import string


def get_all_subclasses(cls):
    '''Returns all subclasses of a given base class'''
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses

def none_or_greater_than(a, b):
	'''returns True if a is greater than b or None'''
	if a:
		return a > b
	else:
		return True

def none_or_less_than(a, b):
	'''returns True if a is less than b or None'''
	if a:
		return a < b
	else:
		return True

def id_generator(size=6, chars=string.ascii_uppercase):
	'''generates random strings of specified lenght within known ranges'''
	return ''.join(random.choice(chars) for _ in range(size))

def first(iterable, default = None):
  '''Just resturns the first item of an iterable. With default.'''
  iterator = iter(iterable)
  return next(iterator, default)